/*
  Title:    awana car rally finish line detector
  Author:   samson
  Date:     mar 12, 2004
  Purpose:  finish line detector outputs to serial port
  needed
  Software: AVR-GCC
  needed
  Hardware: ATS90S2313
*/

#include <io.h>
#include <interrupt.h>
#include <sig-avr.h>
#include <timer.h>
#include <string.h>
#include <pgmspace.h>
#include <eeprom.h>

#include "naked.h"
#include "types.h"
#include "uart.h"

/*****************************************************\
| defines                                             |
\*****************************************************/
// eeprom locations
#define kEEPROM_ID                    0

// how many pins on port B are we using?
#define kNumInputs                    4
#define kDetectorPins                 0x0f

#define kInvalidVal                   0xff

// this is how many counts we have on the time in a second
#define kTimerCountsPerSecond (KCPUFrequency / KClockPrescalar)

// buttons
#define kCommitButtonBit              0x10
#define kResetButtonBit               0x20
#define kNewSelectionButtonBit        0x40
#define kStartTriggerBit              0x80

#define kButtonBits                   0xf0

// button sticky bits are reset by the controller only by the reset function
#define kCommitButtonStickyBit        0x01
#define kResetButtonStickyBit         0x02
#define kNewSelectionButtonStickyBit  0x04
#define kStartTriggerStickyBit        0x08

#define kButtonStickyBits             0x0f

// LEDs
#define kLED0                         2
#define kLED1                         3
#define kLED2                         4
#define kLED3                         5
#define kLEDReady                     6

// ready LED bit flags
#define kReadyLEDBlink                0x01
#define kReadyLEDOn                   0x02

#define kReadyLEDFlags                0x03

// ready LED timer values
#define kReadyLEDUartMaxTimer         255

// commands
#define kCmdRequest                   '>'
#define kCmdBroadcast                 '!'
#define kCmdAddressOK                 '#'
#define kCmdEnd                       '.'
#define kCmdNull                      0

#define kCmdQueryPositions            'P'
#define kCmdQueryTime                 'T'
#define kCmdSetReadyLED               'L'
#define kCmdQueryFlags                'F'
#define kCmdLocalReset                'r'
#define kCmdSetAddress                'A'
#define kCmdAllReset                  'R'

#define kCmdLEDBlink                  'b'
#define kCmdLEDOn                     '1'

#define kStsReply                     '<'
#define kStsEnd                       '.'

#define kStsQueryPositions            'p'
#define kStsQueryTime                 't'
#define kStsSetReadyLED               'l'
#define kStsQueryFlags                'f'
#define kStsLocalReset                'r'

#define kStsFlagStartY                'S'
#define kStsFlagStartN                's'
#define kStsFlagCommitY               'C'
#define kStsFlagCommitN               'c'
#define kStsFlagResetY                'R'
#define kStsFlagResetN                'r'
#define kStsFlagNewSelectionY         'N'
#define kStsFlagNewSelectionN         'n'

/*****************************************************\
| global variables                                    |
\*****************************************************/
// this is the command id of the device
register BYTE theID asm ("r6");

// keep track of the trigger order of the pins
register BYTE thePositionCounter asm ("r7");
BYTE thePositions[kNumInputs];

// array to store results from
// - debug timer is time kNumInputs (last one)
WORD theTimes[kNumInputs];

// uart activity - blink timer (visual only)
register BYTE theUartActivityBlinkTimer asm ("r8");

// current command
register BYTE theCurrentCommand asm ("r9");

// additional input
register BYTE theAdditionalInput asm ("r10");

// buffer button presses
register BYTE theButtonStates asm ("r16");

// ready LED
register BYTE theReadyLEDState asm ("r17");

// use of registers here is probably not correct because
// gcc expects them to be saved, but hopefully, we can
// get away with it because registers are allocated to
// functions top-down.

// mask off each bit as it gets triggered
// - DO __NOT__ USE ASM R1 HERE BECAUSE IT IS THE
//   COMPILER'S ZERO REGISTER
register BYTE theMask asm ("r2");

// high resolution counter (24 bit + 8 bit in the timer itself)
// - 'A' is the least significant byte
register BYTE theTimerA asm ("r3");
register BYTE theTimerB asm ("r4");
register BYTE theTimerC asm ("r5");

// scratch buffer
char theScratchBuf[6];

/*****************************************************\
| extra read-only data                                |
\*****************************************************/

#define XROM __attribute__ ((section ("xrom")))

const char kProgramID0[] XROM;
const char kProgramID0[] = "AWANA";

const char kProgramID1[] XROM;
const char kProgramID1[] = "GRAND PRIX";

/*
const char kProgramID2[] XROM;
const char kProgramID2[] = "FINISH LINE";

const char kProgramID3[] XROM;
const char kProgramID3[] = "DETECTOR";
*/

const char kVersion[] XROM;
const char kVersion[] = "v 2.0";

const char kAuthor[] XROM;
const char kAuthor[] = "Samson Chee";

/*****************************************************\
| inline assembler                                    |
\*****************************************************/
// 24 bit increment: 'A' is the least significant byte
// - assume that r1 is the zero register
#define inc24(pA, pB, pC) __asm__ __volatile__ \
  ("eor r1, r1" "\n\t" \
   "inc r1"     "\n\t" \
   "add %0, r1" "\n\t" \
   "eor r1, r1" "\n\t" \
   "adc %1, r1" "\n\t" \
   "adc %2, r1" "\n\t" \
   :                   \
   "=r" (pA),          \
   "=r" (pB),          \
   "=r" (pC)           \
   :                   \
   : "r1", "cc")

// swap nibbles instruction
#define swapNibbles(b) __asm__ __volatile__ \
  ("swap %0" "\n\t" \
   :                \
   "=r" (b)         \
   : )

// don't care if we clobber everything
#define softreset() __asm__ __volatile__ \
  ("eor r30, r30" "\n\t" \
   "eor r31, r31" "\n\t" \
   "ijmp"         "\n\t" \
   : : :"r30", "r31")

/*****************************************************\
| forward defines                                     |
\*****************************************************/
void resetTimer(void);
DWORD getTimer(void);

void resetPositions(void);

/*****************************************************\
| interrupt handlers                                  |
\*****************************************************/
NAKED(SIG_OVERFLOW0) {
  cli();
  
  // timer 0 overflow
  pushR0();
  pushR1();
  pushMEM(SREG);

  // increment A, B, C (chained)
  // - the physical timer will autoincrment meanwhile as necessary
  // - inc24 uses r1, and this is naked, so need to push and pop r1
  inc24(theTimerA, theTimerB, theTimerC);

  // interrupt return
  popMEM(SREG);
  popR1();
  popR0();

  // interrupt return
  sei();
  reti();
}

/*****************************************************\
| function defintions                                 |
\*****************************************************/

/*****************************************************\
| conversion functions                                |
\*****************************************************/
// number to ascii - no bounds check
inline char numToAsc(BYTE pNum) {
  return (pNum == kInvalidVal) ? '-' : pNum + '0';
}

// single number to single hex char (takes lower nibble only)
char hexChar(BYTE pNum) {
  BYTE theChar = pNum & 0x0f;

  if (theChar < 10) {
    theChar += '0';
  } else {
    theChar += 'A' - 10;
  }

  return theChar;
}

char* WORDToHex(WORD pNum) {
  // read directly off the registers...
  register BYTE b0 asm ("r25");
  register BYTE b1 asm ("r24");

  // r24 is special because it gets passed as the argument
  pushREG(b1);

  // r25 is special because it gets clobbered on return
  pushREG(b0);
  theScratchBuf[1] = hexChar(b0);
  popREG(b0);

  swapNibbles(b0);
  theScratchBuf[0] = hexChar(b0);


  // restore r24 because it just got messed up
  popREG(b1);

  pushREG(b1);
  theScratchBuf[3] = hexChar(b1);
  popREG(b1);

  swapNibbles(b1);
  theScratchBuf[2] = hexChar(b1);
  // end r24 special handling

  theScratchBuf[4] = 0;
  
  return theScratchBuf;
}

// is the character a valid address character?
BYTE isValidAddress(BYTE pChar) {
  if ((pChar >= '0') && (pChar <= '9')) return 1;
  if ((pChar >= 'A') && (pChar <= 'Z')) return 1;
  if ((pChar >= 'a') && (pChar <= 'z')) return 1;
  return 0;
}

/*****************************************************\
| command handlers                                    |
\*****************************************************/
// reset
void reset(void) {
  // setup main polling loop
  // - call resetTimer last for accuracy
  theButtonStates &= ~kButtonStickyBits;
  theReadyLEDState = kReadyLEDOn;
  theCurrentCommand = kCmdNull;

  resetPositions();
  resetTimer();
}

// convert time into 1/1000 seconds
WORD convertTime(DWORD pTicks) {
  WORD theTime;
  theTime = (WORD)(pTicks / (WORD)(KCPUFrequency / (KClockPrescalar * 1000)));

  return theTime;
}

// reset timer
void resetTimer(void) {
  // clear epoch timers and individual timers
  BYTE i;
  
  for (i = 0; i < kNumInputs; i++) {
    theTimes[i] = 0;
  }

  // clear soft timer values
  theTimerA = 0;
  theTimerB = 0;
  theTimerC = 0;

  // clear hard timer value
  outp(0x00, TCNT0);
}

// reset positions
void resetPositions(void) {
  // in case the position doesn't get set, then it will stay being 'last'
  BYTE theInvalidVal = kInvalidVal;

  theMask |= kDetectorPins;
  thePositionCounter = 1;

  thePositions[0] = theInvalidVal;
  thePositions[1] = theInvalidVal;
  thePositions[2] = theInvalidVal;
  thePositions[3] = theInvalidVal;
}

void cmdReply(char pSts, char* pStr) {
  uart_putchar(kStsReply);
  uart_putchar(theID);
  uart_putchar(pSts);
  uart_putstr(pStr);
  uart_putchar(kStsEnd);
}

// command buffer handler
void handleCmdBuf(void) {
  BYTE theChar;

  theChar = uart_getchar();
  if (theChar) {
    // signal uart activity
    theUartActivityBlinkTimer = kReadyLEDUartMaxTimer;

    // process command
    if (theChar == kCmdRequest) {
      // got the attention character
      theCurrentCommand = kCmdRequest;

      // reset additional input
      theAdditionalInput = kInvalidVal;

    } else if (theCurrentCommand == kCmdRequest) {
      // expect address
      if ((theChar == theID) || (theChar == kCmdBroadcast)) {
        theCurrentCommand = kCmdAddressOK;
      } else {
        theCurrentCommand = kCmdNull;
      }

    } else if (theCurrentCommand == kCmdAddressOK) {
      // decode command now
      switch (theChar) {
        case kCmdQueryPositions : 
        case kCmdQueryTime :
        case kCmdSetReadyLED :
        case kCmdQueryFlags :
        case kCmdLocalReset :
        case kCmdSetAddress :
        case kCmdAllReset :
          {
            // these are valid commands
            theCurrentCommand = theChar;
          }; break;
        default :
          {
            theCurrentCommand = kCmdNull;
          }
      } // switch (theChar)

      // reset additional data
      theAdditionalInput = 0;

    } else if (theChar == kCmdEnd) {
      // execute command
      switch (theCurrentCommand) {
        case kCmdQueryPositions : 
          {
            // return positions
            BYTE i;
            for (i = 0; i < kNumInputs; i++) {
              theScratchBuf[i] = numToAsc(thePositions[i]);
            }
            theScratchBuf[kNumInputs] = 0;

            cmdReply(kStsQueryPositions, theScratchBuf);
          }; break;
        case kCmdQueryTime :
          {
            if ((theAdditionalInput >= '1') && (theAdditionalInput <= '4')) {
              WORDToHex(theTimes[theChar - '1']);
              cmdReply(kStsQueryTime, theScratchBuf);
            }
          }; break;
        case kCmdSetReadyLED :
          {
            // reset LED state
            theReadyLEDState &= ~kReadyLEDFlags;

            switch (theAdditionalInput) {
              case kCmdLEDOn :
                {
                  theReadyLEDState |= kReadyLEDOn;
                }; break;
              case kCmdLEDBlink :
                {
                  theReadyLEDState |= kReadyLEDBlink;
                }; break;
            }

            // reply
            cmdReply(kStsSetReadyLED, 0);
          }; break;
        case kCmdQueryFlags :
          {
            // return flag state
            theScratchBuf[0] =
              (theButtonStates & kStartTriggerBit) ? kStsFlagStartY : kStsFlagStartN;
            theScratchBuf[1] =
              (theButtonStates & kCommitButtonBit) ? kStsFlagCommitY : kStsFlagCommitN;
            theScratchBuf[2] =
              (theButtonStates & kResetButtonBit) ? kStsFlagResetY : kStsFlagResetN;
            theScratchBuf[3] =
              (theButtonStates & kNewSelectionButtonBit) ? kStsFlagNewSelectionY : kStsFlagNewSelectionN;
            theScratchBuf[4] = 0;

            // reply
            cmdReply(kStsQueryFlags, theScratchBuf);

            // reset button states
            theButtonStates &= ~kButtonBits;
          }; break;
        case kCmdLocalReset :
          {
            // reset controller state
            reset();
            cmdReply(kStsLocalReset, 0);
          }; break;
        case kCmdSetAddress :
          {
            if (isValidAddress(theAdditionalInput)) {
              // set new id
              theID = theAdditionalInput;

              // store character in eeprom
              cli();
              eeprom_wb(kEEPROM_ID, theID);
            }
          }; // note: no break here because we want to softreset
        case kCmdAllReset :
          {
            // this is the only command where a reply is not necessary
            softreset();
          }; // note: no break here because softreset reinitializes the controller
      } // switch (theChar)

      // finished processing command now
      theCurrentCommand = kCmdNull;

    } else if ((theCurrentCommand != kCmdNull) && (theAdditionalInput == 0)) {
      // if we get to here, we need additional input
      theAdditionalInput = theChar;
    
    } else {
      // can't handle this command...
      // - should actually be (theCurrentCommand == kCmdNull)
      theCurrentCommand = kCmdNull;

    } // if (theChar == kCmdRequest)

  } // if (theChar)
} // handleCmdBuf()

/*****************************************************\
| misc. functions                                     |
\*****************************************************/
// get timer
DWORD getTimer(void) {
  UDWORD_QUADBYTE UDQ;

  // populate the bytes of the union
  UDQ.B[0] = inp(TCNT0);
  UDQ.B[1] = theTimerA;
  UDQ.B[2] = theTimerB;
  UDQ.B[3] = theTimerC;

  return UDQ.D;
}

// main loop
void main_loop(void) {
  // loop variables
  BYTE i, j;

  // what we have in port B
  BYTE thePortIn;

  // main polling loop
  // - loop forever
  while (1) {
    handleCmdBuf();

    // ----------------------------------------------------
    // trigger pins sections
    // - start trigger and race end indicators
    // - pins PB7 are the start signals
    //   - these are not debounced: epoch times don't track
    //     the timer when the start switch is low, so they
    //     effectively start counting the difference in time
    //     as long as the signal is held low
    // - pins PB4, PB5 are the race end indicators
    //   - as soon as these trigger, the race is finished
    // ----------------------------------------------------

    // get port input
    thePortIn = inp(PINB);
    
    // ----------------------------------------------------
    // position pins section
    // ----------------------------------------------------
    BYTE theDetectors = thePortIn & kDetectorPins;

    // see what new pins have triggered
    theDetectors = theDetectors & theMask;
    if (theDetectors) {
      // if the timer hasn't already started, then start it now to record
      // differential times
      if (!(theButtonStates & kStartTriggerStickyBit)) {
        resetTimer();
        theButtonStates |= kStartTriggerStickyBit;
        theReadyLEDState &= ~kReadyLEDOn;
      }

      // if we have something, then record when the event got triggered

      // j is the mask to be shifted
      j = 0x01;
      
      for (i = 0; i < kNumInputs; i++) {
        // check if we're masking the right bit?
        if (j & theDetectors) {

          // - record the position that pin 'i' came in
          // - increment position
          thePositions[i] = thePositionCounter;
          theTimes[i] = convertTime(getTimer());
          thePositionCounter++;
        }

        // iterative step
        j <<= 1;
      } // for (i < kNumInputs);
    } // if (theDetectors)

    // update mask
    theMask &= ~theDetectors;

    // ----------------------------------------------------
    // buttons
    // ----------------------------------------------------
    thePortIn = ~thePortIn;

    // Button states are only set by the main loop.
    // - Button states reset by flag request from serial port
    //   except for the start trigger which is only reset by
    //   the reset button

    if (thePortIn & kCommitButtonBit) {
      theButtonStates |= kCommitButtonBit;
    }

    if (thePortIn & kResetButtonBit) {
      theButtonStates |= kResetButtonBit;
      reset();
    }

    if (thePortIn & kNewSelectionButtonBit) {
      theButtonStates |= kNewSelectionButtonBit;
    }

    if (thePortIn & kStartTriggerBit) {
      if (!(theButtonStates & kStartTriggerStickyBit)) {
        resetTimer();
        theButtonStates |= kStartTriggerStickyBit;
        theReadyLEDState &= ~kReadyLEDOn;
      }

      theButtonStates |= kStartTriggerBit;
    }

    // ----------------------------------------------------
    // display output LEDs
    // ----------------------------------------------------

    // display position data
    WORD theTime = convertTime(getTimer()) / 100;  // time in 1/10 seconds

    if ((theTime % thePositions[0]) == 0) {
      sbi(PORTD, kLED0);
    } else {
      cbi(PORTD, kLED0);
    }

    if ((theTime % thePositions[1]) == 0) {
      sbi(PORTD, kLED1);
    } else {
      cbi(PORTD, kLED1);
    }

    if ((theTime % thePositions[2]) == 0) {
      sbi(PORTD, kLED2);
    } else {
      cbi(PORTD, kLED2);
    }

    if ((theTime % thePositions[3]) == 0) {
      sbi(PORTD, kLED3);
    } else {
      cbi(PORTD, kLED3);
    }

    // display ready light
    if (theReadyLEDState & kReadyLEDBlink) {
      // Blink
      if ((theTime % 8) <= 3) {
        sbi(PORTD, kLEDReady);
      } else {
        cbi(PORTD, kLEDReady);
      }

    } else if (theReadyLEDState & kReadyLEDOn) {
      // On
      if (theUartActivityBlinkTimer == 0) {
        cbi(PORTD, kLEDReady);
      } else {
        sbi(PORTD, kLEDReady);
      }

    } else {
      // Off (default state)
      if (theUartActivityBlinkTimer == 0) {
        sbi(PORTD, kLEDReady);
      } else {
        cbi(PORTD, kLEDReady);
      }

    }

    if (theUartActivityBlinkTimer) {
      theUartActivityBlinkTimer--;
    }

  } // while (1)
} // main_loop()

/*****************************************************\
| program entry point                                 |
\*****************************************************/
int main( void ) {
  // get my id from eeprom first...
  theID = eeprom_rb(kEEPROM_ID);

  /*
    port b used for input
    - b.0 to b.3 for photo-detector module, b.4 to b.7 for buttons
    - continuously poll the port to determine ordering
    - on reset, mask is all cleared
      - as each pin is triggered, mask is set to include that pin
    - we are finished when all pins are triggered or if the
      external stop signal is received on port b.4 or b.5

    detector in
      in PB.0 polling from detector module logic outputs 
      in PB.1   
      in PB.2   
      in PB.3   

    --------------------------------------------------------------------------------
 
    buttons
      in PB.4 commit race 
      in PB.5 reset race 
      in PB.6 new selection 
    trigger
      in in PB.7 starts race timer (optional) 

    --------------------------------------------------------------------------------
 
    serial data
      in PD.0 serial data receive 
      out PD.1 serial data transmit 

    --------------------------------------------------------------------------------
 
    output LEDs
      out PB.2 outputs race results 
      out PB.3   
      out PB.4   
      out PB.5   

    --------------------------------------------------------------------------------
 
    ready light
      out PD.6 ready to race 
  */

  /* debug only
  outp(0xff, DDRB);
  int i, j;
  for (i = 0; i < 255; i++) {
    for (j = 0; j < 32000; j++) {
      outp(i, PORTB);
    }
  }
  */


  // setup port B
  outp(0x00, DDRB);  // data direction to input
  outp(0xf0, PORTB); // pull port b.4 to b.7 high (buttons)


  // setup port D
  outp(0x7e, DDRD);  // data direction to output except for serial.rx
  outp(0x00, PORTD); // don't source current when pulled low


  // now set up special function ports

  // set timer prescaler
  outp(KClockPrescalarCode, TCCR0);

  // enable interrupts
  // - only enable timer 0 for now
  timer_enable_int(0x02);

  // init uart
  uart_init();

  // initialize
  reset();

  // enable interrupts
  sei();

  // main loop
  main_loop();

  // done?  but we shouldn't ever get here...
  softreset();
  return 0;
}

